/**
 * @(#)tsid.cpp 2000/09/25
 *
 * @author Ice
 * @version 1.02
 *
 * ICE Team free software group
 *
 * This file is part of TSID.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _TSID
#define _TSID

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include "tsid.h"

/**
 * Construnct the TSID
 */
TSID::TSID() {
  WDEBUG("Construct the TSID\n");

  versionString="tsid 1.02 by Ice Team Free Software Group (C) 2000-2001\n";

  // base directory is not directly visible
  TDir=".tsid_hvsc";

  HVSC_DIR="HVSC_DIR";

  lastError=NO_TSID_ERROR;
  homeDir=NULL;
  homeTDir=NULL;
  actualAbsFilePath=NULL;
  bufferAbsFilePath=NULL;
  pSidt=NULL;

  // read the home directory of user
  if (!getHomeDir()) {
    cerr << "Error: not home directory found\n";
  }

  if (!setHomeTDir()) {
    cerr << "Error";
  }

  WDEBUG("home dir= " << homeDir << endl);
  WDEBUG("home TSID dir= " << homeTDir << endl);
}

/**
 * Destruct the TSID
 */
TSID::~TSID() {
  WDEBUG("Destruct the TSID\n");

  flush();

  delete(homeDir);
  delete(homeTDir);
  delete(actualAbsFilePath);
  delete(bufferAbsFilePath);
}

/**
 * Set the absolute path to HVSC directory.
 *
 * @param pathToHVSC absolute path to HVSC directory
 * @return true if the operation is ok
 */
bool TSID::setBaseDir(const char *pathToHVSC) {
  WDEBUG("Set Base Directory to: " << pathToHVSC << endl);

  TSID::pathToHVSC=new char[strlen(pathToHVSC)+1];
  if (TSID::pathToHVSC==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;     // no memory available
      return false;
  }
  strcpy(TSID::pathToHVSC, pathToHVSC);
  return true;
}

/**
 * Set the absolute path to HVSC directory using enviroment
 *
 * @return true if operation ok
 */
bool TSID::setBaseDir() {
  WDEBUG("Set base directory using the enviroment: " << HVSC_DIR << endl);

  char *tmp=getenv(HVSC_DIR);
  if (tmp==NULL) {
    cerr << "No enviroment for " << HVSC_DIR << " set up\n";
    lastError=NO_ENV;                        // no enviromnt
    return false;
  }

  return setBaseDir(tmp);
}

/**
 * Add the listen time to the corrisponding file in THVSC
 *
 * @param time the number of listened second
 * @param tuneNumber the number of tune being listen
 * @param hvscAbsFilePath file path absolute to hvsc base directory
 * @return true if the operation is ok
 */
bool TSID::addTime(int time, int tuneNumber, const char *hvscAbsFilePath) {
  WDEBUG("Add " << time << " second to " << hvscAbsFilePath <<
         " tune #" << tuneNumber << endl);

  if ((pathToHVSC==NULL) || (strlen(pathToHVSC)==0)) {
    WDEBUG ("Skip operation due to invalid path\n");
    lastError=INVALID_PATH;    // invalid path
    return false;
  }

  char *subStr=strstr(hvscAbsFilePath, pathToHVSC);

  // if the relative path is not found, exit
  if (subStr==NULL) {
    WDEBUG ("Skip operation due to invalid path\n");
    lastError=INVALID_PATH;    // invalid path
    return false;
  }

  // relative name pointer
  char *relName=subStr+strlen(pathToHVSC);

  WDEBUG("Relative name= " << relName << endl);

  addTime(relName, time, tuneNumber);
  return true;
}

/**
 * Add the listen time to the corrisponding file in TSID HVSC
 *
 * @param hvscRelFilePath file path relative to hvsc base directory
 * @param time the number of listened second
 * @param tuneNumber the number of tune being listen
 * @return true if the operation is ok
 */
bool TSID::addTime(const char *hvscRelFilePath, int time, int tuneNumber) {
  #define check \
    char *str=getStrBeforeSlash(actualAbsFilePath); \
    if (!checkDir(str)) { \
      lastError=DIR_CREATE; \
      cerr << "Error checking/creating directory: " << str << endl; \
      return false; \
    }

  WDEBUG("Add " << time << " second to " << hvscRelFilePath <<
         " tune #" << tuneNumber << endl);

  lastError=NO_TSID_ERROR;

  // skip if low listen time
  if (time < MIN_LISTEN_TIME) {
    WDEBUG("Skip operation due to low time\n");
    return true;
  }

  if (strlen(hvscRelFilePath)==0) {
    WDEBUG ("Skip operation due to invalid path\n");
    lastError=INVALID_PATH;    // invalid path
    return false;
  }

  // calculate absolute path of the file
  actualAbsFilePath=new char[strlen(homeTDir)+strlen(hvscRelFilePath)+2];
  if (actualAbsFilePath==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;     // no memory available
      return false;
  }
  sprintf(actualAbsFilePath, "%s%st", homeTDir, hvscRelFilePath);

  WDEBUG("Absolute path is: "<< actualAbsFilePath << endl);

  if (bufferAbsFilePath!=NULL) {
    if (strcmp(actualAbsFilePath, bufferAbsFilePath)==0) {
      // add the information to current song
      pSidt->add(tuneNumber, time, 1);
    } else {

        if (!flush()) {                      // flush information
          return false;
        }

        check;                               // check directory

        pSidt=new SIDT(actualAbsFilePath);   // use a new sidt for this song
        pSidt->add(tuneNumber, time, 1);     // add information
        delete bufferAbsFilePath;
        bufferAbsFilePath=actualAbsFilePath; // remember this new song name
      }
  } else {
      check;

      pSidt=new SIDT(actualAbsFilePath);     // use a new sidt for this song
      pSidt->add(tuneNumber, time, 1);       // add information
      bufferAbsFilePath=actualAbsFilePath;   // remember this new song name
    }
  return true;
}

/**
 * Flush the information not saved to file
 *
 * @return true if operation is ok
 */
bool TSID::flush() {
  WDEBUG("Flush the information to file\n");

  lastError=NO_TSID_ERROR;

  if (pSidt!=NULL) {
    if (!pSidt->read()) {                    // add just present information
      lastError=READ_FILE;
      return false;
    }
    if (!pSidt->write()) {                   // write collected information
      lastError=WRITE_FILE;
      return false;
    }
    delete pSidt;
  }
  return true;
}

/**
 * Get home directory of user
 * The private variables homeDir is chaged
 * Note: this code is taken and modified from XMame!!!
 *
 * @return true if operation is ok
 */
bool TSID::getHomeDir() {
  struct passwd *pw;

  lastError=NO_TSID_ERROR;

  pw=getpwuid(getuid());
  if(!pw) {
      cerr << "Who are you? Not found in passwd database!!\n";
      lastError=NO_USER;  // no user information
      return false;
  }

  homeDir=new char[strlen(pw->pw_dir)+1];
  if (homeDir==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;  // no memory available
      return false;
  }
  strcpy(homeDir, pw->pw_dir);

  return true;
}

/**
 * Set the home directory of TSID tree
 * If the directory doesn't exists it will be created
 * The private variables homeTDir is chaged
 *
 * @return true if operation is ok
 */
bool TSID::setHomeTDir() {
  int handle;               // handle for file
  int err;                  // error is not 0 in file operation

  lastError=NO_TSID_ERROR;

  homeTDir=new char[strlen(homeDir)+strlen(TDir)+2];
  if (homeTDir==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;  // no memory available
      return false;
  }
  sprintf(homeTDir, "%s/%s",homeDir, TDir);

  // verify if base directory tree is present
  if ((handle=open(homeTDir, O_RDONLY))==-1) {
                                              // not present, must create it
    err=mkdir(homeTDir, O_CREAT);
    err+=chmod(homeTDir, S_IRUSR | S_IWUSR | S_IEXEC);

    WDEBUG("Creating directory :\""<<homeTDir<<"\"\n");
    if (err) {                                // error creating directory ?
      lastError=DIR_CREATE;                   // error creating directory
      return false;
    }
  } else {                                    // present, close handle
      close(handle);
    }

  return true;
};

/**
 * Get the string before the slash of a passed string
 *
 * @param str the passed string
 * @return the expected string
 */
char *TSID::getStrBeforeSlash(char *str) {
  int length;                                // length of result string
  char *pnt=&str[strlen(str)];               // pointer to the end of string

  while (*pnt!=PATH_SEP) {                   // look for slash
    pnt--;                                   // decrement pointer
    if (pnt<=str) break;                     // avoid endless loop
  }

  char *result=new char[(length=(pnt-str))+1];
  if (result==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;                   // no memory available
      return NULL;
  }
  strncpy(result, str, length);

  result[length]='\0';                       // terminate the string

  WDEBUG("BeforeSlash: \""<< result << "\"\n");
  return result;
}

/**
 * Check if the path is present, otherwise creates it.
 *
 * @param path the path to check
 * @return true if the generation/check of path is ok
 */
bool TSID::checkDir(char *path) {
  char *subPath;            // sub string of path
  int handle;               // handle for file
  int err;                  // error is not 0 in file operation
  bool result=true;         // return result

  lastError=NO_TSID_ERROR;

  // verify if base directory tree is present
  if ((handle=open(path, O_RDONLY))==-1) {
                                              // not present, must create it
    subPath=getStrBeforeSlash(path);

    result=checkDir(subPath);                 // check the parent directory
    delete(subPath);
                                              // create the subdir
    err=mkdir(path, O_CREAT);
    err+=chmod(path, S_IRUSR | S_IWUSR | S_IEXEC);

    WDEBUG("Creating directory: \"" << path << "\"\n");
    if (err) {                                // there's error?
      lastError=DIR_CREATE;                   // error creating directory
      result=false;                           // yes
    }

  } else {                                    // present, close handle
      close(handle);
    }

  return result;
}

#endif // _TSID.CPP


